/*
 * ifdepd -- interface dependency daemon
 *
 * Copyright (c) 2005 Alexander Hausner <alex@hugo.bmg.gv.at>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *      $Id: ifdepd.c,v 1.0 2005/02/14 14:47:34 alh Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <syslog.h>
#include <stdarg.h>
#include <sys/event.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/if_media.h>
#include <net/route.h>

#define BUF_SIZE 2048
#define STR_DEL ":"

#define CHR_DEL STR_DEL[0]
#define POLL_INTERVAL 2

#define VERSION "Version 1.01 21022005"

typedef struct sig_rec {
  int signal;
  sig_t function;
} sig_array[];

typedef const char *int_name;

static int_name *s_ints=NULL, *d_ints=NULL;
static int l_socket, kq, verbose = 0, quiet = 0, d = 0;
#ifndef POLLING_MODE
static int r_socket;
static struct kevent kev;
#endif

void Perror(const char *cmd)
{
  switch (errno) {
  case ENXIO:
    if (d) {
      syslog(LOG_CRIT, "%s: no such interface", cmd);
      exit(1);
    } else {
      errx(1, "%s: no such interface", cmd);
    }
    break;
  case EPERM:
    if (d) {
      syslog(LOG_CRIT, "%s: permission denied", cmd);
      exit(1);
    } else {
      errx(1, "%s: permission denied", cmd);
    }
    break;
  default:
    if (d) {
      syslog(LOG_CRIT, "%s", cmd);
      exit(1);
    } else {
      err(1, "%s", cmd);
    }
  }
}

int int_status(int_name if_name) {
  struct ifmediareq ifmr;
  int ret_value;

  (void) memset(&ifmr, 0, sizeof ifmr);
  (void) strncpy(ifmr.ifm_name, if_name, sizeof ifmr.ifm_name);
  if (ioctl(l_socket, SIOCGIFMEDIA, (caddr_t)&ifmr) < 0) {
    if (d) {
      syslog(LOG_CRIT, "ioctl");
      exit(1);
    } else {
      err(1, "ioctl");
    }
  }
  ret_value = ifmr.ifm_status & IFM_ACTIVE;
  if (verbose) {
    if (d) {
      (ret_value) ? syslog(LOG_CRIT, "%s status: up", if_name) :
	syslog(LOG_CRIT, "%s status: down", if_name);
    } else {
      (ret_value) ? warnx("%s status: up", if_name) :
	warnx("%s status: down", if_name);
    }
  }
  return ret_value;
}
  
void int_up(int_name if_name) {
  struct ifreq ifr;
  int flags;

  (void) memset(&ifr, 0, sizeof ifr);
  ifr.ifr_addr.sa_family = AF_INET;
  strncpy(ifr.ifr_name, if_name, sizeof ifr.ifr_name);
  if (ioctl(l_socket, SIOCGIFFLAGS, (caddr_t)&ifr) < 0)
    Perror("ioctl (SIOCGIFFLAGS)");
  flags = (ifr.ifr_ifru.ifru_flags[0] & 0xffff) | (ifr.ifr_ifru.ifru_flags[1] << 16);
  flags |= IFF_UP;
  ifr.ifr_ifru.ifru_flags[0] = flags & 0xffff;
  ifr.ifr_ifru.ifru_flags[1] = flags >> 16;
  if (ioctl(l_socket, SIOCSIFFLAGS, (caddr_t)&ifr) < 0)
    Perror("ioctl (SIOCSIFFLAGS)");
  else if (verbose) {
    if (d) {
      syslog(LOG_CRIT, "%s up", if_name);
    } else {
      warnx("%s up", if_name);
    }
  }
}

void int_down(int_name if_name) {
  struct ifreq ifr;
  int flags;

  (void) memset(&ifr, 0, sizeof ifr);
  ifr.ifr_addr.sa_family = AF_INET;
  strncpy(ifr.ifr_name, if_name, sizeof ifr.ifr_name);
  if (ioctl(l_socket, SIOCGIFFLAGS, (caddr_t)&ifr) < 0)
    Perror("ioctl (SIOCGIFFLAGS)");
  flags = (ifr.ifr_ifru.ifru_flags[0] & 0xffff) | (ifr.ifr_ifru.ifru_flags[1] << 16);
  flags &= ~IFF_UP;
  ifr.ifr_ifru.ifru_flags[0] = flags & 0xffff;
  ifr.ifr_ifru.ifru_flags[1] = flags >> 16;
  if (ioctl(l_socket, SIOCSIFFLAGS, (caddr_t)&ifr) < 0)
    Perror("ioctl (SIOCSIFFLAGS)");
  else if (verbose) {
    if (d) {
      syslog(LOG_CRIT, "%s down", if_name);
    } else {
      warnx("%s down", if_name);
    }
  }
}

inline void cleanup(int_name *s_ints, int_name *d_ints) {
  if (s_ints != NULL)
    free(s_ints);
  if (d_ints != NULL)
    free(d_ints);
}

void usage(const char *progname) {
  errx(1, "usage: %s [-v|q] [-d] -S interface[:interface...] -D interface[:interface...]",
       progname);
}

inline int del_count(const char *s, const char d) {
  int cnt=0, a;
  for (a=0; s[a] != '\0'; a++)
    if (s[a] == d)
      cnt++;
  return cnt;
}

int_name *strtarry(char *str, const int cnt) {
  int_name *ret_value;
  int a;

  if ((ret_value=(int_name *) malloc(sizeof(int_name)*cnt)) == NULL)
    err(1, NULL);

  for(a=0; a<cnt; a++)
    ret_value[a]=strsep(&str, STR_DEL);

  return ret_value;
}

int ints_exists(int_name *ints, int cnt) {
  for(;--cnt>=0;)
    if (if_nametoindex(ints[cnt]) == 0) {
      if (d) {
	syslog(LOG_CRIT, "interface %s does not exist", ints[cnt]);
      } else {
	warnx("interface %s does not exist", ints[cnt]);
      }
      return 0;
    }
  return 1;
}

int s_int_ok(int_name *ints, int cnt) {
  for(;--cnt>=0;)
    if (!int_status(ints[cnt])) {
      if (verbose) {
	if (d) {
	  syslog(LOG_CRIT, "%s failed", ints[cnt]);
	} else {
	  warnx("%s failed", ints[cnt]);
	}
      }
      return 0;
    }
  return 1;
}

void next_event(void){
#ifndef POLLING_MODE
  char msg[BUF_SIZE];

  if (kevent(kq, NULL, 0, &kev, 1, NULL) == -1) {
    if (d) {
      syslog(LOG_CRIT, "error get kevent");
      exit(1);
    } else {
      err(1, "error get kevent");
    }
  }
  while (read(r_socket, msg, sizeof msg) == sizeof msg);
#else
  (void) sleep(POLL_INTERVAL);
#endif
}

void state_ok(int_name *s_ints, int s_cnt,
	      int_name *d_ints, int d_cnt) {
  time_t ok_time;
  if (!quiet) {
    ok_time = time(NULL);
    if (d) {
      syslog(LOG_CRIT, "%s    --> state_ok", ctime(&ok_time));
    } else {
      warnx("%s    --> state_ok", ctime(&ok_time));
    }
  }
  for(;--d_cnt>=0;)
    int_up(d_ints[d_cnt]);
  while (s_int_ok(s_ints, s_cnt)) {
    next_event();
    if (verbose) {
      if (d) {
	syslog(LOG_CRIT, "got kevent in state_ok");
      } else {
	warnx("got kevent in state_ok");
      }
    }
  }
}

void state_fail(int_name *s_ints, int s_cnt,
		int_name *d_ints, int d_cnt) {
  time_t fail_time;
  if (!quiet) {
    fail_time = time(NULL);
    if (d) {
      syslog(LOG_CRIT, "%s    --> state_fail", ctime(&fail_time));
    } else {
      warnx("%s    --> state_fail", ctime(&fail_time));
    }
  }
  for(;--d_cnt>=0;)
    int_down(d_ints[d_cnt]);
  while (!s_int_ok(s_ints, s_cnt)) {
    next_event();
    if (verbose) {
      if (d) {
	syslog(LOG_CRIT, "got kevent in state_fail");
      } else {
	warnx("got kevent in state_fail");
      }
    }
  }
}

void automat(int_name *s_ints, int s_cnt,
	     int_name *d_ints, int d_cnt) {
  while(1) {
    state_ok(s_ints, s_cnt, d_ints, d_cnt);
    state_fail(s_ints, s_cnt, d_ints, d_cnt);
  }
}

void done(int act_signal) {
  time_t fail_time;

  if (d)
    syslog(LOG_CRIT, "got signal: %d", act_signal);
  else
    warnx("got signal: %d", act_signal);
  cleanup(s_ints, d_ints);
  close(l_socket);
#ifndef POLLING_MODE
  close(r_socket);
#endif
  if (d) {
    closelog();
  }
  if (!quiet) {
    fail_time = time(NULL);
    if (d)
      syslog(LOG_CRIT, "%s    --> end", ctime(&fail_time));
    else
      warnx("%s    --> end", ctime(&fail_time));
  }
  exit(0);
}

int reg_signals(void) {
  int a;
  sig_array sa = {
    {SIGINT, &done},
    {SIGHUP, SIG_IGN},
    {SIGPIPE, &done},
    {SIGALRM, &done},
    {SIGXCPU, &done},
    {SIGXFSZ, &done},
    {SIGVTALRM, &done},
    {SIGTERM, &done}
  };

  for(a=0; a<(sizeof(sa)/sizeof(struct sig_rec)); a++)
    if (signal(sa[a].signal, sa[a].function) == SIG_ERR)
      return 0;
  return 1;
}

int main(int argc, char *argv[]) {
  int S=0, D=0;
#ifndef POLLING_MODE
  int off=0;
#endif
  const char *switches = "vqdS:D:";
  char ch;
  int s_cnt=0, d_cnt=0;
  time_t fail_time;
#ifndef POLLING_MODE
  struct timespec ts={0,0};
#endif

  if (!reg_signals())
    errx(1, "reg_signals() error");

  if (argc < 4)
    usage(argv[0]);

  while ((ch = getopt(argc, argv, switches)) != -1)
    switch(ch) {
    case 'd':
      d = 1;
      break;
    case 'v':
      if (quiet)
	usage(argv[0]);
      verbose = 1;
      break;
    case 'q':
      if (verbose)
	usage(argv[0]);
      quiet = 1;
      break;
    case 'S':
      S = 1;
      s_cnt = del_count(optarg, CHR_DEL)+1;
      s_ints = strtarry(optarg, s_cnt);
      if (!ints_exists(s_ints, s_cnt)) {
	free(s_ints);
	if (D)
	  free(d_ints);
	exit(1);
      }
      break;
    case 'D':
      D = 1;
      d_cnt = del_count(optarg, CHR_DEL)+1;
      d_ints = strtarry(optarg, d_cnt);
      if (!ints_exists(d_ints, d_cnt)) {
	free(d_ints);
	if (S)
	  free(s_ints);
	exit(1);
      }
      break;
    default:
      usage(argv[0]);
    }
  
  if (! (S && D))
    usage(argv[0]);

  if (d && rfork(RFNOWAIT|RFPROC|RFFDG)) {
    cleanup(s_ints, d_ints);
    return 0;
  }
 
  if ((kq = kqueue()) == -1) {
    cleanup(s_ints, d_ints);
    if (d) {
      syslog(LOG_CRIT, "kqueue");
      exit(1);
    } else {
      err(1, "kqueue");
    }
  }
#ifndef POLLING_MODE
  if ((r_socket = socket(PF_ROUTE, SOCK_RAW, AF_UNSPEC)) < 0) {
    cleanup(s_ints, d_ints);
    if (d) {
      syslog(LOG_CRIT, "routing socket");
      exit(1);
    } else {
      err(1, "routing socket");
    }
  }
  if (setsockopt(r_socket, SOL_SOCKET, SO_USELOOPBACK,
		 &off,sizeof(off)) < 0) {
    cleanup(s_ints, d_ints);
    close(r_socket);
    if (d) {
      syslog(LOG_CRIT, "setsockopt SO_USELOOPBACK");
      exit(1);
    } else {
      err(1, "setsockopt SO_USELOOPBACK");
    }
  }
#endif
  if ((l_socket = socket(PF_INET, SOCK_RAW, AF_UNSPEC)) < 0) {
    cleanup(s_ints, d_ints);
#ifndef POLLING_MODE
    close(r_socket);
#endif
    if (d) {
      syslog(LOG_CRIT, "link socket");
      exit(1);
    } else {
      err(1, "link socket");
    }
  }
#ifndef POLLING_MODE
  EV_SET(&kev, r_socket, EVFILT_READ, EV_ADD|EV_ENABLE|EV_CLEAR, 0, 0, NULL);
  if (kevent(kq, &kev, 1, NULL, 0, &ts) == -1) {
    if (d) {
      syslog(LOG_CRIT, "register kevent");
      exit(1);
    } else {
      err(1, "register kevent");
    }
  }
#endif
  if (d) {
    openlog(argv[0], LOG_PID | LOG_ODELAY, LOG_DAEMON);
  }
  if (!quiet) {
    fail_time = time(NULL);
    if (d) {
      syslog(LOG_CRIT, VERSION);
      syslog(LOG_CRIT, "%s    --> start", ctime(&fail_time));
    } else {
      warnx(VERSION);
      warnx("%s    --> start", ctime(&fail_time));
    }
  }
  automat(s_ints, s_cnt, d_ints, d_cnt);
  // never reached
  cleanup(s_ints, d_ints);
  close(l_socket);
#ifndef POLLING_MODE
  close(r_socket);
#endif
  return 1;
}
